package fi.aalto.control.server;

import java.util.Locale;

import com.prosysopc.ua.stack.builtintypes.LocalizedText;
import com.prosysopc.ua.stack.builtintypes.NodeId;
import com.prosysopc.ua.stack.core.Identifiers;

import com.prosysopc.ua.StatusException;
import com.prosysopc.ua.nodes.UaNodeFactoryException;
import com.prosysopc.ua.nodes.UaNode;
import com.prosysopc.ua.server.NodeManagerUaNode;
import com.prosysopc.ua.server.UaInstantiationException;
import com.prosysopc.ua.server.UaServer;
import com.prosysopc.ua.server.nodes.PlainVariable;
import com.prosysopc.ua.server.nodes.UaObjectNode;
import com.prosysopc.ua.server.nodes.UaVariableNode;

public class ControlNodeManager extends NodeManagerUaNode {

	public static final String NAMESPACE = "http://localhost/OPCUA/ControlAddressSpace";
	UaVariableNode input;
	UaVariableNode output;

	public ControlNodeManager(UaServer arg0, String arg1) {
		super(arg0, arg1);
		// Auto-generated constructor stub
	}

	private void createAddressSpace() throws StatusException, UaInstantiationException {
		// Create demo address space

		int ns = getNamespaceIndex();
		final UaNode objectsFolder = getServer().getNodeManagerRoot().getObjectsFolder();

		UaNode parent = objectsFolder;

		// Base objects
		parent = createFolder(ns, "CoDeSys.OPC.DA", parent);
		parent = createFolder(ns, "PLC_GW3", parent);
		parent = createFolder(ns, "Application", parent);
		parent = createFolder(ns, "GVL_WP_HPP", parent);
				
		// Input
		input = createVariable(ns, "P300_MeasVal", new Float(0), Identifiers.Float, parent);
		((ControlHistoryManagerListener)getHistoryManager().getListener()).addVariableHistory(input);
		
		// Output
		output = createVariable(ns, "M200_CurCtrlVal", new Float(0), Identifiers.Float, parent);
		((ControlHistoryManagerListener)getHistoryManager().getListener()).addVariableHistory(output);
		
		// PIC300 Pressure PID control
		createVariable(ns, "PIC300_CtrlOff", false, Identifiers.Boolean, parent);
		createVariable(ns, "PIC300_CtrlOn", true, Identifiers.Boolean, parent);
		//createVariable(ns, "PIC300_CtrlVal", 0, Identifiers.Float, parent);
		//createVariable(ns, "PIC300_CurSPVal", new Float(60), Identifiers.Float, parent);
		createVariable(ns, "PIC300_Enable", true, Identifiers.Boolean, parent);
		createVariable(ns, "PIC300_Kp", new Float(1), Identifiers.Float, parent);
		//createVariable(ns, "PIC300_Td", new Float(0), Identifiers.Float, parent);
		//createVariable(ns, "PIC300_Ti", new Float(0), Identifiers.Float, parent);
		createVariable(ns, "PIC300_PIDReset", false, Identifiers.Boolean, parent);
		createVariable(ns, "PIC300_SPVal", new Float(60), Identifiers.Float, parent);

		return;
	}	

	@Override
	protected void init() throws StatusException, UaNodeFactoryException {
		super.init();

		createAddressSpace();
	}

	private UaObjectNode createFolder(int ns, String name, UaNode parent)
	{
	    final NodeId id = new NodeId(ns, name);
	    UaObjectNode node = new UaObjectNode(this, id, name, Locale.ENGLISH);
	    node.setTypeDefinitionId(Identifiers.FolderType);
		try {
		    this.addNodeAndReference(parent, node, Identifiers.Organizes);
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}
		node.setDescription(new LocalizedText(name)); 
		
		return node;
	}

	private <T> PlainVariable<T> createVariable(int ns, String name, T value, NodeId dataTypeId, UaNode parent)
	{
		NodeId id = new NodeId(ns, name);
		PlainVariable<T> variable = new PlainVariable<T>(this, id, name, LocalizedText.NO_LOCALE);
		variable.setDataTypeId(dataTypeId);
		variable.setTypeDefinitionId(Identifiers.BaseDataVariableType);
		parent.addComponent(variable);
		variable.setCurrentValue(value);
		variable.setDescription(new LocalizedText(name)); 

		return variable;
	}
	
}
